import numpy as np


## \brief   Create list of ECDF values.
#
#  This function creates the empirical cumulative distribution functions from a list of distances
#  and a list of bin values. That is, each element of the resulting list tells how many elements of
#  the distance list are smaller than the respective bin value.
#
#  \param   distance_list  List of distances to be grouped into the bins.
#  \param   bins           List of bins.
#  \retval  ecdf_list      Resulting list of amout of distances that are smaller than resp. bins.
def empirical_cumulative_distribution_vector( distance_list, bins ):
  return [ np.sum( [distance < basket for distance in distance_list] ) / len(distance_list) \
           for basket in bins ]  # np.sum appears to be much faster than Python's standard sum!

## \brief   Assemble matrix of (generalized) distances between elemenst of datasets.
#
#  This function creates a matrix whose (i,j)th entry corresponds to the distance between element
#  i of a subset of dataset_a and element j of a subset of dataset_b. The respective subsets are
#  characterized by the indices of the respective first and last elements.
#  Notably, the matrix entries can be numbers or more general data types (such as lists).
#
#  \param   dataset_a      First dataset, whose subset is compared to second dataset.
#  \param   dataset_b      Second dataset, whose subset is compared to first dataset.
#  \param   distance_fct   Function generating a generalized distance between members of datasets.
#  \param   start_a        Starting index of considered subset of dataset_a. Defaults to 0. 
#  \param   end_a          Last (exclusive) index of consideres subset. Defaults to len(dataset).
#  \param   start_b        Starting index of considered subset of dataset_b. Defaults to 0. 
#  \param   end_b          Last (exclusive) index of consideres subset. Defaults to len(dataset).
#  \retval  distance_mat   Matrix of generalized distances.
def create_distance_matrix( dataset_a, dataset_b, distance_fct, 
  start_a=0, end_a=None, start_b=0, end_b=None ):
  if end_a is None:  end_a = len(dataset_a)
  if end_b is None:  end_b = len(dataset_b)

  if end_a < start_a or end_b < start_b:
    raise Exception("Invalid subset indices chosen.")

  return [ [ distance_fct(dataset_a[i], dataset_b[j]) for j in range(start_b, end_b) ] \
             for i in range(start_a, end_a) ]


## \brief   Assemble ecdf vector, whose elements are list of values for all subset combinations.
#
#  This function assembles a list of ecdf vectors for all possible combinations of subsets of the
#  dataset. Importantly, none of the subsets are compared to themselves and subsets i and j are
#  compared only once (not i with j and j with i).
#  The first dimension of the result refers to the index of the bin / ecdf vector. The second index
#  of the result refers to the subset combination.
#
#  \param   dataset        First dataset, whose subset are compared to one another.
#  \param   bins           List of bins.
#  \param   distance_fct   Function generating a generalized distance between members of dataset.
#  \param   subset_indices List of starting (and ending) indices of disjointly subdivided dataset.
#  \retval  ecdf_list      ecdf vector enlisting values for subset combinations.
def empirical_cumulative_distribution_vector_list( dataset, bins, distance_fct, subset_indices ):
  if not all(subset_indices[i] <= subset_indices[i+1] for i in range(len(subset_indices)-1)):
    raise Exception("Subset indices are out of order.")
  if subset_indices[0] != 0 or subset_indices[-1] != len(dataset):
    raise Exception("Not all elements of the dataset are distributed into subsets.")

  matrix = []
  for i in range(len(subset_indices)-1):
    for j in range(i):
      distance_list = create_distance_matrix(dataset, dataset, distance_fct, 
        subset_indices[i], subset_indices[i+1], subset_indices[j], subset_indices[j+1])
      while isinstance(distance_list[0], list):
        distance_list = [item for sublist in distance_list for item in sublist]
      matrix.append( empirical_cumulative_distribution_vector(distance_list, bins) )

  return np.transpose(matrix)

## \brief   Same as empirical_cumulative_distribution_vector_list, but for bootstrapping.
#
#  \param   dataset_a      First dataset, whose elements are compared to second dataset.
#  \param   dataset_b      Second dataset, whose elements iare compared to first dataset.
#  \param   bins           List of bins.
#  \param   distance_fct   Function generating a generalized distance between members of dataset.
#  \param   n_samples      Number of perturbatins of the datasets.
#  \retval  ecdf_list      ecdf vector enlisting values for subset combinations.
def empirical_cumulative_distribution_vector_list_bootstrap(
  dataset_a, dataset_b, bins, distance_fct, n_samples ):
  distance_matrix = np.array( create_distance_matrix(dataset_a, dataset_b, distance_fct) )
  matrix = []
  for _ in range(n_samples):
    permute_a = np.random.randint(distance_matrix.shape[0], size=distance_matrix.shape[0])
    permute_b = np.random.randint(distance_matrix.shape[1], size=distance_matrix.shape[1])
    distance_list = np.ndarray.flatten( distance_matrix[permute_a,permute_b] )
    matrix.append( empirical_cumulative_distribution_vector(distance_list, bins) )
  return np.transpose(matrix)


## \brief  Vector of means of the ecdf vectors.
#
#  \param   ecdf_list      Usually the result of empirical_cumulative_distribution_vector_list.
#  \retval  ecdf_means     Element-wise mean values of the ecdf vectors.
def mean_of_ecdf_vectors( ecdf_vector_list ):
  return [ np.mean(vector) for vector in ecdf_vector_list ]

## \brief  Covariance matrix of ecdf vectors.
#
#  \param   ecdf_list      Usually the result of empirical_cumulative_distribution_vector_list.
#  \retval  ecdf_covar     Covariance matrix associated to the ecdf vectors.
def covariance_of_ecdf_vectors( ecdf_vector_list ):
  return np.cov( ecdf_vector_list )


## \brief  Evaluate target/objective/cost function associated to estimator type from dataset.
#
#  Evaluate the negative log-likelihood in the way that is characterized by the estimator.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   dataset        The dataset with respect to which the target function is evaluated.
#  \retval  target_val     The value of the target function.
def evaluate( estimator, dataset ):
  return estimator.evaluate( dataset )

## \brief  Evaluate target/objective/cost function associated to estimator type from ecdf vector.
#
#  Evaluate the negative log-likelihood in the way that is characterized by the estimator.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   ecdf_vector    The vector of ecdf, which is the argument for the target function.
#  \retval  target_val     The value of the target function.
def evaluate_from_empirical_cumulative_distribution_functions( estimator, vector ):
  mean_deviation = np.subtract( estimator.mean_vector , vector )
  if not estimator.error_printed:
    try:
      return np.dot( mean_deviation , np.linalg.solve(estimator.covar_matrix, mean_deviation) )
    except np.linalg.LinAlgError as error:
      estimator.error_printed = True
      print("WARNING: Covariance matrix is singular. CIL_estimator uses different topology.")
  return np.dot( mean_deviation , mean_deviation )



import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import chi2
import ecdf_estimator.utils as ecdf_aux


## \brief  Plot all ecdf vectors.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   plotter        Python object to which the plot should be added. Defaults to plt.
#  \param   plot_options   String describing the options for plotting.
#  \retval  plotter        Python object to whhich the plot has been added.
def plot_ecdf_vectors( estimator, plotter=plt, plot_options="b." ):
  if hasattr(estimator, 'bins'):  bins = estimator.bins
  else:                           bins = range(1, len(estimator.ecdf_list)+1)
  
  for vector in np.transpose(estimator.ecdf_list):
    plotter.plot(bins, vector, plot_options)
  return plotter


## \brief  Plot means of ecdf vectors.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   plotter        Python object to which the plot should be added. Defaults to plt.
#  \param   plot_options   String describing the options for plotting.
#  \retval  plotter        Python object to whhich the plot has been added.
def plot_mean_vector( estimator, plotter=plt, plot_options="g." ):
  if hasattr(estimator, 'bins'):  bins = estimator.bins
  else:                           bins = range(1, len(estimator.ecdf_list)+1)
  
  plotter.plot(bins, estimator.mean_vector, plot_options)
  return plotter


## \brief  Plot chi square test.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   plotter        Python object to which the plot should be added. Defaults to plt.
#  \param   plot_options   String describing the options for plotting.
#  \retval  plotter        Python object to whhich the plot has been added.
def plot_chi2_test( estimator, plotter=plt, plot_options="r-" ):
  n_logl = [ ecdf_aux.evaluate_from_empirical_cumulative_distribution_functions(estimator, vector) \
             for vector in np.transpose(estimator.ecdf_list) ]
  khi, bins = np.histogram( n_logl )
  khi_n = [ x / sum(khi) / (bins[1] - bins[0]) for x in khi ]
  plotter.hist(bins[:-1], bins, weights=khi_n)
  df = len( estimator.ecdf_list )
  x  = np.linspace(chi2.ppf(0.01, df), chi2.ppf(0.99,df), 100)
  plotter.plot(x, chi2.pdf(x, df),plot_options, lw=5, alpha=0.6, label='chi2 pdf')
  return plotter


## \brief  Save ecdf vectors, mean of ecdf vectors, covariance matrix and bin values to files.
#
#  \param   estimator      The estimator class defining the specifics of the target function.
#  \param   name           Prefix of the file names to which data is saved.
def save_data( estimator, name="ecdf_estimator" ):
  np.savetxt(name + '_ecdf-list.txt',    estimator.ecdf_list, fmt='%.6f')
  np.savetxt(name + '_mean-vector.txt',  estimator.mean_vector, fmt='%.6f')
  np.savetxt(name + '_covar-matrix.txt', estimator.covar_matrix, fmt='%.6f')
  if hasattr(estimator, "bins"):  np.savetxt(name + '_bins.txt', estimator.bins, fmt='%.6f')



import numpy as np
import ecdf_estimator.utils as ecdf_aux


## \brief  Objective function assembled in the stardard ecdf way.
class standard:
  ## \brief  Construct objective function.
  def __init__( self, dataset, bins, distance_fct, subset_sizes ):
    self.dataset        = dataset
    self.bins           = bins
    self.distance_fct   = distance_fct
    self.subset_indices = [ sum(subset_sizes[:i]) for i in range(len(subset_sizes)+1) ]
    self.ecdf_list      = ecdf_aux.empirical_cumulative_distribution_vector_list(
                            dataset, bins, distance_fct, self.subset_indices )
    self.mean_vector    = ecdf_aux.mean_of_ecdf_vectors(self.ecdf_list)
    self.covar_matrix   = ecdf_aux.covariance_of_ecdf_vectors(self.ecdf_list)
    self.error_printed  = False

  def evaluate_ecdf(self, dataset):
    comparison_set = np.random.randint( len(self.subset_indices)-1 )
    distance_list = ecdf_aux.create_distance_matrix(self.dataset, dataset,
      self.distance_fct, self.subset_indices[comparison_set], self.subset_indices[comparison_set+1])
    while isinstance(distance_list[0], list):
      distance_list = [item for sublist in distance_list for item in sublist]
    return ecdf_aux.empirical_cumulative_distribution_vector(distance_list, self.bins)

  def evaluate( self, dataset ):
    return ecdf_aux.evaluate_from_empirical_cumulative_distribution_functions( self,
      self.evaluate_ecdf(dataset) )


## \brief  Objective function assembled via bootstrapping.
class bootstrap:
  ## \brief  Construct objective function.
  def __init__( self, dataset_a, dataset_b, bins, distance_fct, n_samples=1000 ):
    self.dataset_a      = dataset_a
    self.dataset_b      = dataset_b
    self.bins           = bins
    self.distance_fct   = distance_fct
    self.n_samples      = n_samples
    self.ecdf_list      = ecdf_aux.empirical_cumulative_distribution_vector_list_bootstrap(
                            dataset_a, dataset_b, bins, distance_fct, self.n_samples )
    self.mean_vector    = ecdf_aux.mean_of_ecdf_vectors(self.ecdf_list)
    self.covar_matrix   = ecdf_aux.covariance_of_ecdf_vectors(self.ecdf_list)
    self.error_printed  = False

  def evaluate_ecdf( self, dataset ):
    if np.random.randint( 2 ) == 0:  comparison_set = self.dataset_a
    else:                            comparison_set = self.dataset_b

    distance_list = ecdf_aux.create_distance_matrix(comparison_set, dataset, self.distance_fct)
    distance_list = [item for sublist in distance_list for item in sublist]
    return ecdf_aux.empirical_cumulative_distribution_vector(distance_list, self.bins)

  def evaluate( self, dataset ):
    return ecdf_aux.evaluate_from_empirical_cumulative_distribution_functions( self,
      self.evaluate_ecdf(dataset) )


## \brief  Objective function that consists of mutliple objective functions.
class multiple:
  ## \brief  Construct objective function.
  def __init__( self, obj_fun_list ):
    self.obj_fun_list = obj_fun_list

    n_rows, n_columns = 0, -1
    for obj_fun in obj_fun_list:
      n_rows += obj_fun.ecdf_list.shape[0]
      if n_columns == -1:
        n_columns = obj_fun.ecdf_list.shape[1]
      elif n_columns != obj_fun.ecdf_list.shape[1]:
        print("ERROR: All objective functions should contain the same number of ecdf vectors.")

    self.ecdf_list = np.zeros( (n_rows, n_columns) )
    index = 0
    for obj_fun in obj_fun_list:
      self.ecdf_list[index:index+obj_fun.ecdf_list.shape[0],:] = obj_fun.ecdf_list
      index = index+obj_fun.ecdf_list.shape[0]

    self.mean_vector    = ecdf_aux.mean_of_ecdf_vectors(self.ecdf_list)
    self.covar_matrix   = ecdf_aux.covariance_of_ecdf_vectors(self.ecdf_list)
    self.error_printed  = False

  def evaluate( self, dataset ):
    vector = [ obj_fun.evaluate_ecdf(dataset) for obj_fun in self.obj_fun_list ]
    while isinstance(vector[0], list):
      vector = [item for sublist in vector for item in sublist]
    return ecdf_aux.evaluate_from_empirical_cumulative_distribution_functions( self, vector )



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
  long_description = fh.read()

setuptools.setup(
  name='ecdf_estimator',
  version='0.1.3',
  author='Andreas Rupp',
  author_email='info@rupp.ink',
  description='Python package for parameter estimation of random data',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='https://github.com/AndreasRupp/ecdf_estimator',
  project_urls = {
    "Bug Tracker": "https://github.com/AndreasRupp/ecdf_estimator/issues"
  },
  license='LGPL-2.1',
  packages=['ecdf_estimator'],
  install_requires=[
    'requests',
    'numpy>=1.19.5',
    'matplotlib>=3.6'
  ],
)



import numpy as np
import ecdf_estimator.utils as ecdf_aux


## \brief  Heuristically determine region in which useful bin/radii values are located.
#
#  \param   dataset_a        Subset of the overall dataset.
#  \param   dataset_b        Different subset of the poverall dataset.
#  \param   distance_fct     Functions that evaluates (generalized) distance between subset members.
#  \param   rel_offset       Relative offset to determin interval of reasonable bin values.
#  \param   rel_cutoff       Relative cutoff of the interval of reasonable bin values.            
#  \retval  target_val       The value of the target function.
def estimate_radii_values( dataset_a, dataset_b, distance_fct, rel_offset=0.05, rel_cutoff=0.05 ):
  distance_data = [ distance_fct(data_a, data_b) for data_b in dataset_b for data_a in dataset_a ]
  while isinstance(distance_data[0], list):
    distance_data = [item for sublist in distance_data for item in sublist]
  distance_data = np.sort(distance_data)

  data_offset = round(len(distance_data) * rel_offset)
  radius_max  = distance_data[-(data_offset + 1)]
  radius_min  = distance_data[data_offset]

  upper_bound = radius_max + rel_cutoff * (radius_max - radius_min)
  lower_bound = radius_min - rel_cutoff * (radius_max - radius_min)
  if lower_bound < 0:
    lower_bound = radius_min

  return lower_bound, upper_bound, distance_data


## \brief  Heuristically determine reasonable bin/radii values from larger choice.
#
#  \param   distance_list    List of distances to be grouped into the bins.
#  \param   bins             List of possible bin values.
#  \param   n_bins           Maximum amount of bins, which are to be selected. Defaults to 10.
#  \param   choose_typ       Heurustic that is used to select bins.
#  \param   min_value_shift  Exclude values that are smaller than min value of distances plus this.
#  \param   max_value_shift  Exclude values that are largr than max value of distances plus this.
#  \retval  target_val       The value of the target function.
def choose_bins(distance_list, possible_bins, n_bins=10, choose_type="uniform_y_dist",
  min_value_shift=None, max_value_shift=None ):
  ecdf_curve = ecdf_aux.empirical_cumulative_distribution_vector(distance_list, possible_bins)
  if choose_type == "uniform_y_dist":
    max_value, min_value = np.amax( ecdf_curve ), np.amin( ecdf_curve )
    if min_value_shift is None:  min_value_shift = (max_value - min_value) / n_bins
    if max_value_shift is None:  max_value_shift = (min_value - max_value) / n_bins
    step_size = ( (max_value+max_value_shift) - (min_value+min_value_shift) ) / n_bins
    indices = []
    for _ in range(n_bins):
      if not indices:  index = np.argmax( ecdf_curve >= min_value + min_value_shift )
      else:            index = np.argmax( ecdf_curve >= ecdf_curve[indices[-1]] + step_size )

      if ecdf_curve[index] > max_value+max_value_shift:  break
      indices.append( index )
    return [ possible_bins[i] for i in indices ]
  elif choose_type == "uniform_y":
    max_value, min_value = np.amax( ecdf_curve ), np.amin( ecdf_curve )
    if min_value_shift is None:  min_value_shift = (max_value - min_value) / n_bins
    if max_value_shift is None:  max_value_shift = (min_value - max_value) / n_bins
    rad_bdr   = np.linspace( min_value+min_value_shift , max_value+max_value_shift , num=n_bins )
    indices   = [ np.argmax( ecdf_curve >= bdr ) for bdr in rad_bdr ]
    indices   = [ index for index in indices if ecdf_curve[index] <= max_value+max_value_shift ]
    unique_indices = np.unique(indices)
    if len(indices) != len(unique_indices):
      print("WARNING: Some bins were duplicate. These duplicates are removed from the list.")
    return [ possible_bins[i] for i in unique_indices ]
  elif choose_type == "uniform_x":
    max_index, min_index = np.amax( np.argmin(ecdf_curve) ), np.amin( np.argmax(ecdf_curve) )
    if min_value_shift is None:  min_value_shift = (max_index - min_index) / n_bins
    if max_value_shift is None:  max_value_shift = (min_index - max_index) / n_bins
    indices   = np.linspace( min_index+min_value_shift , max_index+max_value_shift , num=n_bins )
    unique_indices = np.unique(indices)
    if len(indices) != len(unique_indices):
      print("WARNING: Some bins were duplicate. These duplicates are removed from the list.")
    return [ possible_bins[int(i)] for i in unique_indices ]
  else:
    print("WARNING: Invalid choose_type flag for choose_bins. Nothing is done in this function.")
